/*
 * Josh Lifton and Michael Broxton
 * MIT Media Lab
 * Copyright (c) 2002-2004. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
Be sure to "erase code space" before uploading the bootloader!

Power up.
Blink status LED twice.  Blink all three colors of RGB LED.
Set RGB LED blue if there is no good OS and wait indefinitely.  If there is an OS, keep RGB LED off.
Set OS wait timer to 1 minute.
Turn blue on if instructed to wait indefinitely for an OS.
Turn red on whenever a transmission begins.  Blue and green are off.
Toggle blue whenever a byte that is not ignored is received.
Turn off LEDs after receipt of a new OS or if an OS was ignored.
Turn back blue if an OS failed to receive correctly and waiting indefinitely.
After correctly receiving or ignoring OS and if waiting indefinitely, turn blue-green and wait forever.

*/


#include <c8051F000.h>
#include <PushpinHardwareV3.h>
#include <PushpinBootloader.h>

// The data structure containing all necessary information about a current OS.
typedef struct OSStateDefinition {
	unsigned int randomSeed;
	unsigned int size;
	unsigned int version;
	unsigned char isOld;
	unsigned char isGood;
	unsigned char crc;
	unsigned char stack;
} OSState;

// The data structure containing all necessary information about an incoming OS.
typedef struct OSHeaderDefinition {
	unsigned int size;
	unsigned int version;
	unsigned char crc;
	unsigned char stack;
} OSHeader;  

// Flash memory macros.
#define enableFlashErase() PSCTL = 0x03
#define disableFlashErase() PSCTL = 0x00
#define enableFlashWrite() PSCTL = 0x01
#define disableFlashWrite() PSCTL = 0x00
// UART macros.
#define setUARTMode1() SM0 = 0; SM1 = 1
// Timer macros.
#define enableTimer1() TR1 = 1
#define disableTimer1() TR1 = 0
// General interrupt macros.
#define disableGlobalInterrupts() EA = 0
/*
 * This is the reload values for timer 1 needed to generate
 * the appropriate baud rate for communication over the IR
 * channel.  This is the maximum baud rate that should be used
 * when incoming bytes are written to flash memory.
 */
#define BAUD92160 0xF1		// 10kOhm, 1085pF
#define BAUDRATE TH1	// The timer 1 reload register generates baud rate.


/*
 * Function prototypes.
 */
unsigned char listenForNewOS(void);
void processByte(unsigned char incomingByte);
void configureIRCommModule(void);
unsigned char getFlashBlockCRC8(unsigned char code *, unsigned int);
void clearFlashMemory(unsigned char code *, unsigned int);
void saveOSState(OSState xdata *);
void copyXdataToFlash(unsigned char code *, unsigned char xdata *, unsigned int);
void configurePins(void);
void disableWatchdogTimer(void);
void enableExternalClock(void);
void initializeHardware(void);
void delay(unsigned int);
void initializeOSState(void);

/*
 * Global variables.
 */
extern const unsigned char code CRC8Table[256];
extern const unsigned char code decodeNybbleTable[256];
const code at MINI_PAGE_ADDRESS OSState osState = 0x00;
unsigned char xdata serialState;
unsigned char data byteCount;
OSHeader xdata osHeader;
OSState xdata tempOSState;
unsigned char xdata * data currentBytePtr;
unsigned char data resetVectorByte0;
unsigned char data resetVectorByte1;
unsigned char data resetVectorByte2;
unsigned char data redundancyBuffer[NUMBER_OF_RETRANSMITS];
unsigned char xdata shouldWaitForNewOS;

// meme : probably shouldn't be global, but wanted to use the IDE to watch them.
unsigned char xdata byteBuffer;
unsigned char data byteToCopy;

// meme : shouldn't need this, but sdcc seems buggy.
unsigned char code * data codePointer;

enum SerialStates {
  WAITING_FOR_REQUEST, WAITING_FOR_END_REQUEST, WAITING_FOR_PACKET_TYPE,
  WAITING_FOR_OSHEADER, IGNORING_OS, RECEIVING_OS, OS_RECEIVE_SUCCESS, 
  // IGNORING_RESET_VECTOR, RECEIVING_INTERRUPT_VECTORS, 
  OS_RECEIVE_FAILURE
};


void main(void) {
	void (* osEntryPoint) () = (void(code *)(void)) OS_CODE_ADDRESS;

	initializeHardware();
  initializeOSState();

  statusLED = 1;
  redRGB = 1;
  delay(0xffff);
  delay(0xffff);
  statusLED = 0;
  redRGB = 0;
  greenRGB = 1;
  delay(0xffff);
  delay(0xffff);
  statusLED = 1;
  delay(0xffff);
  delay(0xffff);
  statusLED = 0;
  greenRGB = 0;
  blueRGB = 1;
  delay(0xffff);
  delay(0xffff);
  blueRGB = 0;

  shouldWaitForNewOS = FALSE;

  /* 
   * If we don't have a good OS, we should wait for one indefinitely.
   */
  if (!(osState.isGood)) {
    blueRGB = 1;
    shouldWaitForNewOS = TRUE;
  }
  while ((listenForNewOS() == FALSE) && shouldWaitForNewOS) {
    statusLED = 0;
    redRGB = 0;
    greenRGB = 0;
    blueRGB = 1;
	}
  if (osState.isGood) {
    if (shouldWaitForNewOS) {
      redRGB = 0;
      greenRGB = 1;
      blueRGB = 1;
      while(1);
    }
		disableTimer1();
    redRGB = 0;
    greenRGB = 0;
    blueRGB = 0;
		SP = osState.stack;
		(*osEntryPoint)();
	}
	else {
		while(1);
	}	
}

unsigned char listenForNewOS(void) {

	unsigned long xdata time;

  bit nybbleRemaining;

  /*
   * Setup comm system.
   */
  configureIRCommModule();
  RI = 0;
  TI = 0;
	REN = 1;
	enableTimer1();

	/*
	 * Listen for incoming bytes.
	 */
  nybbleRemaining = FALSE;
  serialState = WAITING_FOR_REQUEST;
	for (time=TIMETOLISTEN; time != 0; time--) {
    if (serialState == OS_RECEIVE_SUCCESS) {
      return TRUE;
    }
    if (serialState == OS_RECEIVE_FAILURE) {
      shouldWaitForNewOS = TRUE;
      return FALSE;
    }
		if (RI) {
			RI = 0;
      if (!nybbleRemaining) {
        byteBuffer = decodeNybbleTable[SBUF]; 
        nybbleRemaining = TRUE;
      }
      else {
        byteBuffer |= (decodeNybbleTable[SBUF] << 4);
        nybbleRemaining = FALSE;
        processByte(byteBuffer);
        if (serialState == RECEIVING_OS) {
          time = OS_BYTE_TIMEOUT;  // Don't want to time out when receiving an OS.
        }
      }
    }
  }
  if (!(osState.isGood)) {
    shouldWaitForNewOS = TRUE;
  }
  if (serialState == IGNORING_OS) {
    return TRUE;
  }
  return FALSE;
}

void processByte(unsigned char incomingByte) {
  if (serialState == RECEIVING_OS) {
    statusLED = ~statusLED;
    blueRGB = ~blueRGB;
    redundancyBuffer[byteCount++] = incomingByte;
    if (byteCount == NUMBER_OF_RETRANSMITS) {
      byteCount = 0;
      /*
       * Find the most likely byte.
       */
      if ((redundancyBuffer[0] == redundancyBuffer[1]) && 
          (redundancyBuffer[1] == redundancyBuffer[2])) {
        byteToCopy = redundancyBuffer[0];
      }
      else if (redundancyBuffer[0] == redundancyBuffer[1]) { 
        byteToCopy = redundancyBuffer[0];
      }
      else if (redundancyBuffer[0] == redundancyBuffer[2]) {
        byteToCopy = redundancyBuffer[0];
      }
      else if (redundancyBuffer[1] == redundancyBuffer[2]) {
        byteToCopy = redundancyBuffer[1];
      }
      else {
        byteToCopy = redundancyBuffer[2];
      }
      /*
       * Write the choosen byte to flash.
       */
    	enableFlashWrite();
    	*(currentBytePtr) = byteToCopy;
    	disableFlashWrite();
      currentBytePtr++;
      /*
       * Check if the end of the OS has been reached.
       */
      if (currentBytePtr == (unsigned char xdata *) (osHeader.size + ((unsigned int) OS_CODE_ADDRESS))) { 
        statusLED = 0;
        blueRGB = 0;
        if (getFlashBlockCRC8(OS_CODE_ADDRESS, osHeader.size) == osHeader.crc) {
          /*
           * Relocate the interrupt vector table while maintaining the original
           * bootloader reset interrupt vector.
           */
  				resetVectorByte0 = *((unsigned char code *) BOOTLOADER_ADDRESS);
  				resetVectorByte1 = *((unsigned char code *) (BOOTLOADER_ADDRESS + 1));
  				resetVectorByte2 = *((unsigned char code *) (BOOTLOADER_ADDRESS + 2));
  				clearFlashMemory(0, INTERRUPT_VECTOR_TABLE_SIZE);
  				enableFlashWrite();
  				*((unsigned char xdata *) 0) = resetVectorByte0;
  				*((unsigned char xdata *) 1) = resetVectorByte1;
  				*((unsigned char xdata *) 2) = resetVectorByte2;
  				disableFlashWrite();
          codePointer = (unsigned char code *) 3; // meme
          for (currentBytePtr = (unsigned char xdata *) 3; ((unsigned int) currentBytePtr) < INTERRUPT_VECTOR_TABLE_SIZE; currentBytePtr++) {
            byteToCopy = *((unsigned char code *)(((unsigned int) codePointer) + OS_CODE_ADDRESS));
            codePointer++;
            enableFlashWrite();
            *currentBytePtr = byteToCopy;
            disableFlashWrite();
          }

        	/*
        	 * Successfully received a new OS.  Mark it as good.
        	 */
        	tempOSState.randomSeed = osState.randomSeed;
        	tempOSState.isOld = 0;
        	tempOSState.isGood = 1;
        	tempOSState.size = osHeader.size; // Includes the interrupt vectors.
        	tempOSState.version = osHeader.version;
        	tempOSState.crc = osHeader.crc;
        	tempOSState.stack = osHeader.stack;
        	saveOSState(&tempOSState);
        	serialState = OS_RECEIVE_SUCCESS;
        }
        else {
        	/*
        	 * New OS didn't pass the CRC.
        	 */
        	serialState = OS_RECEIVE_FAILURE;
        }
      }
    }
  }   
  else if (serialState == WAITING_FOR_REQUEST && incomingByte == REQUEST) {
    serialState = WAITING_FOR_END_REQUEST;
  }
  else if (serialState == WAITING_FOR_END_REQUEST) {
    if (incomingByte == END_REQUEST) {
      serialState = WAITING_FOR_PACKET_TYPE;
    }
    else {
      serialState = WAITING_FOR_REQUEST;
    }
  }
  else if (serialState == WAITING_FOR_PACKET_TYPE) {
    if (incomingByte == OS_HEADER) {
      serialState = WAITING_FOR_OSHEADER;
      byteCount = 0;
    }
    else if (incomingByte == DONT_WAIT_FOR_OS) {
      serialState = OS_RECEIVE_SUCCESS;
    }
    else if (incomingByte == WAIT_FOR_OS) {
      shouldWaitForNewOS = TRUE;
      serialState = OS_RECEIVE_FAILURE;
    }
  }
  else if (serialState == WAITING_FOR_OSHEADER) {
    /*
     * Record the incoming OS header.
     */
    statusLED = ~statusLED;
    blueRGB = ~blueRGB;
    redRGB = 1;
    greenRGB = 0;
    *(((unsigned char xdata *) (&osHeader)) + byteCount++) = incomingByte;
    if (byteCount == sizeof(OSHeader)) {
			/*
			 * Check that the incoming OS is of the correct size and version.
			 */
			if (osHeader.size < FREE_FLASH_SIZE && (osHeader.version > osState.version || osState.version == 0xFFFF)) {
        statusLED = 0;
        blueRGB = 0;
				/*
				 * Mark the OS as bad until it is verified as good.
				 */
				tempOSState.randomSeed = osState.randomSeed;
				tempOSState.isOld = 0;
				tempOSState.isGood = 0;
				tempOSState.size = osState.size; // Includes the interrupt vectors.
				tempOSState.version = osState.version;
				tempOSState.crc = osState.crc;
				tempOSState.stack = osState.stack;
				saveOSState(&tempOSState);
				/*
				 * Prepare to record the incoming OS.  
				 */
				clearFlashMemory(OS_CODE_ADDRESS, osHeader.size);
        serialState = RECEIVING_OS;
        currentBytePtr = (unsigned char xdata *) OS_CODE_ADDRESS;
        byteCount = 0;
      }
			else {
				/*
				 * Wrong version or size.  Ignore the incoming OS.
				 */
        blueRGB = 0;
        serialState = IGNORING_OS;
        currentBytePtr = 0;
        byteCount = 0;
			}
    }
  }
  else if (serialState == IGNORING_OS) {
    redundancyBuffer[byteCount++] = incomingByte;
    if (byteCount == NUMBER_OF_RETRANSMITS) {
      byteCount = 0;
      currentBytePtr++;
      if (currentBytePtr == (unsigned char xdata *) osHeader.size) {
        serialState = OS_RECEIVE_SUCCESS;
      }
    }
	}
}


/*
 * This function configures the processing module for use 
 * with the IR communication module.  Adapted from detectCommModule().
 */
void configureIRCommModule(void) {
  setUARTMode1();		// 8-bit serial communication.
  T2CON = 0x01;			// Configure timer 2 (no auto-reload).
  CKCON |= 0x20;		// Set timer 2 to run off system clock.
  TMOD |= 0x20;			// Set timer1 as 8-bit auto reload.
  PCON |= 0x80;			// Double baud rate; ie, SMOD = 1.
  PRT1CF &= 0x0F;		//|
  IRChannel0 = 1;		//| 
  IRChannel1 = 1;		//| Set pins P1.[4-7] as inputs.
  IRChannel2 = 1;		//|
  IRChannel3 = 1;		//|
  BAUDRATE = BAUD92160;         // Must agree with the resistor and capacitor values on the IR comm module.
}


/* 
 * Adapted from: http://cell-relay.indiana.edu/mhonarc/cell-relay/1999-Jan/msg00074.html
 * 8 bit CRC Generator, MSB shifted first
 * Polynom: x^8 + x^2 + x^1 + 1
 * 
 * Calculates an 8-bit cyclic redundancy check sum for a block of code 
 * stored in flash memory.  Starts at the first byte in the block and 
 * proceeds to the last byte.
 */
unsigned char getFlashBlockCRC8(unsigned char code * blockPtr, unsigned int blockSize) {
	unsigned int i;
	unsigned char xdata crc = 0;
	for (i=0; i < blockSize; i++) {	
		crc = CRC8Table[crc ^ (*(((unsigned char code *) blockPtr) + i))];
	}
	return crc;
}

/*
 * If the arguments to this function specify a valid block of flash 
 * memory, the smallest area of erasable flash containing that 
 * block will be cleared.  Since the smallest unit of erasable 
 * flash is 512 bytes, up to 1022 extra bytes of memory might be 
 * erased (511 bytes before the block and 511 bytes after the block.
 * Note the the 0x7C00 block of flash cannot be erased by software; 
 * it can only be erased via a JTAG connection.
 */
void clearFlashMemory(unsigned char code * startPtr, unsigned int size) {
	unsigned char interruptState = EA;
  unsigned char xdata * startPtr_xdata = (unsigned char xdata *) startPtr;
	/*
	 * Check that the size is not too large.
	 */
	if (size == 0 || size > FLASH_SIZE || ((unsigned int) startPtr) + size < ((unsigned int) startPtr)) {
		return;
	}
	/*
	 * Erase all but the last flash page within range.
	 */
	while (size > FLASH_PAGE_SIZE) {
		disableGlobalInterrupts();
		enableFlashErase();
    *startPtr_xdata = 0xFF;
		disableFlashErase();
		EA = interruptState;
		size -= FLASH_PAGE_SIZE;
		startPtr += FLASH_PAGE_SIZE;
    startPtr_xdata += FLASH_PAGE_SIZE;
	}
	/*
	 * Erase the last flash page within range.
	 */
	disableGlobalInterrupts();
	enableFlashErase();
  *startPtr_xdata = 0xFF;
	disableFlashErase();
	EA = interruptState;
}

void initializeOSState(void) {
  unsigned char i;
  for (i=0; i<sizeof(OSState); i++) {
    if (*(((unsigned char code *) &osState) + i) != 0xFF) {
      return;
    }
  }
  tempOSState.randomSeed = RANDOM_SEED;
	tempOSState.size = 0x0000;
	tempOSState.version = 0xFFFF;
	tempOSState.isOld = 0x00;
	tempOSState.isGood = 0x00;
	tempOSState.crc = 0x00;
	tempOSState.stack = 0x00;
  saveOSState(&tempOSState);
}

void saveOSState(OSState xdata * state) {
	clearFlashMemory((unsigned char code *) &osState, sizeof(OSState));
	copyXdataToFlash((unsigned char code *) &osState, (unsigned char xdata *) state, sizeof(OSState));
}

/* 
 * Copies a contiguous block of xdata memory to flash memory.  
 * Disables interrupts during sensitive tasks and then restores interrupt state.
 */
void copyXdataToFlash(unsigned char code * to, unsigned char xdata * from, unsigned int size) {
	unsigned int i;
	unsigned char data c;
	unsigned char interruptState = EA;
  unsigned char xdata * data to_xdata = (unsigned char xdata *) to;
	if ((to + size < (unsigned char code *) FLASH_SIZE) && (from + size < (unsigned char xdata *) XDATA_SIZE)) {
		disableGlobalInterrupts();
		for (i=0; i<size; i++) {
			c = *(from + i);
			enableFlashWrite();
      *to_xdata = c;
			disableFlashWrite();
			to++;
      to_xdata++;
		}
		EA = interruptState;
	}			
}

void configurePins(void) {
	PRT0CF = 0xFF;	// Set all of port 0 to push-pull outputs.
	PRT1CF = 0xFF;	// Set all of port 1 to push-pull outputs.
	PRT2CF = 0xFF;	// Set all of port 2 (not pinned out) to push-pull outputs.
	XBR0 |= 0x04;	// UART TX and RX on P0.0 and P0.1.	
	XBR2 |= 0x40;	// Enable crossbar. 
	XBR2 |= 0x80;	// Disable weak pull-ups.
}

void disableWatchdogTimer(void) {
	WDTCN = 0xde;	// This register must be set first.
	WDTCN = 0xad;	// This register must be set second.
}

void enableExternalClock(void) {
	CKCON |= 0x10;				// Don't divide by 12 for base clock; ie, T1M=1.
	OSCXCN = 0x67;				// Start external oscillator > 6.74MHz (using a 22.118MHz crystal)
	delay(0xffff);				// Wait more than a millisecond.
	while (!(OSCXCN & 0x80)) {} // Wait until external oscillator is stable.
	OSCICN |= 0x80;				// Check for missing clock.
	OSCICN |= 0x08;				// Use external crystal.
	FLSCL = ((FLSCL & 0xF0) | 0x09);	// Set flash scalar for 22Mhz clock.
}

void initializeHardware(void) {
	disableWatchdogTimer();
	enableExternalClock();
	configurePins();
	statusLED = 0;
  redRGB = 0;
  greenRGB = 0;
  blueRGB = 0;
}


/* 
 * A blocking function that simply waits an amount of time roughly 
 * proportional to the value of the argument.
 */
void delay(unsigned int interval) {
	while(interval--);
}

//void ISR0(void) interrupt 0 {}
//void ISR1(void) interrupt 1 {statusLED = ~statusLED;}
//void ISR2(void) interrupt 2 {}
//void ISR3(void) interrupt 3 {}
//void ISR4(void) interrupt 4 {}
//void ISR5(void) interrupt 5 {}
//void ISR6(void) interrupt 6 {}
//void ISR7(void) interrupt 7 {}
//void ISR8(void) interrupt 8 {}
//void ISR9(void) interrupt 9 {}
//void ISR10(void) interrupt 10 {}
//void ISR11(void) interrupt 11 {}
//void ISR12(void) interrupt 12 {}
//void ISR13(void) interrupt 13 {}
//void ISR14(void) interrupt 14 {}
//void ISR15(void) interrupt 15 {}
//void ISR16(void) interrupt 16 {}
//void ISR17(void) interrupt 17 {}
//void ISR18(void) interrupt 18 {}
//void ISR19(void) interrupt 19 {}
//void ISR20(void) interrupt 20 {}
//void ISR21(void) interrupt 21 {}
//void ISR22(void) interrupt 22 {}
//void ISR23(void) interrupt 23 {}
//void ISR24(void) interrupt 24 {}
//void ISR25(void) interrupt 25 {}
//void ISR26(void) interrupt 26 {}
//void ISR27(void) interrupt 27 {}
//void ISR28(void) interrupt 28 {}
//void ISR29(void) interrupt 29 {}
//void ISR30(void) interrupt 30 {}
//void ISR31(void) interrupt 31 {}
